# http://talks.golang.org/2012/concurrency.slide#36
# Timeout for whole conversation using select

from csp import timeout, stop, alts

from .boring import boring

def main():
    c = boring("Joe")
    t = timeout(5)
    while True:
        value, chan = yield alts([c, t])
        if chan is c:
            print value
        else:
            print "You talk too much."
            yield stop()



# http://talks.golang.org/2012/concurrency.slide#39
# Daisy-chain

from csp import Channel, put, take, go


def f(left, right):
    yield put(left, 1 + (yield take(right)))


def main():
    n = 100000
    leftmost = Channel()
    right = leftmost
    left = leftmost
    for i in range(n):
        right = Channel()
        go(f(left, right))
        left = right

    def start(c):
        yield put(c, 1)

    go(start(right))
    print (yield take(leftmost))



# http://talks.golang.org/2012/concurrency.slide#27
# Multiplexing

from csp import Channel, put, take, go

from .boring import boring


def fan_in(input1, input2):
    c = Channel()
    def collect(i):
        while True:
            yield put(c, (yield take(i)))
    go(collect(input1))
    go(collect(input2))
    return c


def main():
    c = fan_in(boring("Joe"), boring("Ann"))
    for i in range(10):
        print (yield take(c))
    print "You are boring; I'm leaving."



from csp.impl import dispatch
from csp.impl.channels import ManyToManyChannel as Channel

# TODO: Check resolution
# TODO: Should be reactor-specific
# TODO: Some optimizations?

def timeout(seconds):
    """Returns a channel that will close after the specified timeout (in
    seconds).
    """
    channel = Channel()
    def done():
        channel.close()
    dispatch.queue_delay(done, seconds)
    return channel



from twisted.internet import reactor

from csp.impl.buffers import RingBuffer

class TwistedDispatcher:

    TASK_BATCH_SIZE = 1024

    def __init__(self):
        self.tasks = RingBuffer(32)
        self.running = False
        self.queued = False

    def run(self, f):
        # TODO: Should there be some sort of limit?
        self.tasks.unbounded_unshift(f)
        self.queue()

    def queue_delay(self, f, seconds):
        reactor.callLater(seconds, f)

    def process_messages(self):
        self.running = True
        self.queued = False
        count = 0
        while True:
            try:
                task = self.tasks.pop()
            except IndexError:
                break

            task()

            # TODO: Maybe stop after a max amount of time rather than
            # a max number of tasks.
            count += 1
            if count >= self.TASK_BATCH_SIZE:
                break

        self.running = False
        if len(self.tasks) > 0:
            self.queue()

    def queue(self):
        if self.queued and self.running:
            return

        self.queued = True
        reactor.callLater(0, self.process_messages)

# TODO: Support other dispatchers (e.g. tornado or other reactors)
dispatcher = TwistedDispatcher()

run = dispatcher.run

queue_delay = dispatcher.queue_delay



from zope.interface import implements

from collections import namedtuple

from csp.impl import dispatch
from csp.impl.interfaces import IHandler
from csp.impl.select import do_alts

Instruction = namedtuple("Instruction", ["op", "data"])


# FIX: This is not efficient, right? Python has no "reify"
class FnHandler:
    implements(IHandler)

    def __init__(self, f):
        self.f = f

    def is_active(self):
        return True

    def commit(self):
        return self.f

# TODO: Shouldn't these 2 (and FnHandler) be in "channels" module?


def put_then_callback(channel, value, callback):
    """Puts a value on the channel, calling the supplied callback when
    done, passing False if the channel was closed, True otherwise.
    """
    result = channel.put(value, FnHandler(callback))
    if result:
        callback(result.value)


def take_then_callback(channel, callback):
    """Takes from the channel, calling the supplied callback with the
    received value when done.
    """
    result = channel.take(FnHandler(callback))
    if result:
        callback(result.value)


# XXX
NONE = object()

class Process:

    def __init__(self, gen, finish_callback = None):
        self.gen = gen
        self.finished = False
        self.finish_callback = finish_callback

    def _continue(self, response):
        # print self, "got", response
        # XXX: Is there be a better way to avoid infinite recursion?
        dispatch.run(lambda: self.run(response))

    # TODO: Hmm
    def _done(self, value):
        if not self.finished:
            # print self, "finished"
            self.finished = True
            if self.finish_callback:
                dispatch.run(lambda: self.finish_callback(value))

    def run(self, response = NONE):
        if self.finished:
            return

        try:
            # print self, "send", response
            if response is NONE:
                instruction = self.gen.next()
            else:
                instruction = self.gen.send(response)
        # Normal termination
        except StopIteration:
            self._done(None)
            return
        # Exceptional termination
        except:
            # TODO: Should we put the exception into the channel instead?
            self._done(None)
            raise

        assert isinstance(instruction, Instruction)

        # Early termination
        if instruction.op == "stop":
            self._done(instruction.data)
            return

        if instruction.op == "put":
            channel, value = instruction.data
            put_then_callback(channel, value, self._continue)
            return

        # TODO: Should we throw if the value is an exception?
        if instruction.op == "take":
            channel = instruction.data
            take_then_callback(channel, self._continue)
            return

        # TODO: Timeout channel instead?
        if instruction.op == "wait":
            seconds = instruction.data
            callback = lambda: self._continue(None)
            dispatch.queue_delay(callback, seconds)
            return

        if instruction.op == "alts":
            operations = instruction.data
            result = do_alts(operations, self._continue)
            if result:
                self._continue(result.value)
            return


def put(channel, value):
    """Puts a value onto the channel.

    Must be used with "yield" inside a go function.

    not_closed = yield put(channel, "value")
    """
    return Instruction("put", (channel, value))


def take(channel):
    """Takes a value from the channel.

    Must be used with "yield" inside a go function.

    value = yield take(channel)
    """
    return Instruction("take", channel)


def wait(seconds):
    """Pauses the current go function.

    Must be used with "yield" inside a go function.

    # "Sleeps" for 0.5 seconds (but does not tie up the thread)
    yield wait(0.5)
    """
    return Instruction("wait", seconds)


# TODO: Re-organize code


def alts(operations):
    """Completes at most one of the specified channel operations. Each
    operation must either be a channel to take from, or a tuple of
    (channel-to-put-onto, value-to-put).

    If more than 1 operation is ready, a non-deterministic choice will
    be made.

    Must be used with "yield" inside a go function.

    value = yield alts([channel, ()])
    """
    return Instruction("alts", operations)


# TODO: Better doc
def stop(value=None):
    """Returns from the current go function with the supplied value.

    It's dual to the "return" keyword in normal functions.

    def proc(early):
        yield wait(0.5)
        if early:
            yield stop("early")

        yield stop("late")

    # Will have "early" put on it
    chan1 = go(proc(True))
    # Will have "late" put on it
    chan2 = go(proc(False))
    """
    return Instruction("stop", value)



from csp import Channel, alts

import csp.impl.operations as op


def main():
    # chan = Channel()
    # out = op.map_from(lambda x: x * 2, chan)

    # def counter(n):
    #     i = n
    #     while True:
    #         yield put(chan, i)
    #         i += 1

    # go(counter(0))

    # while True:
    #     value = yield take(out)
    #     if value > 20:
    #         yield stop(value)
    #     print value

    i = Channel()
    def m(x):
        y = x * 2
        print "\t\t\t", x, "=>", y
        return y
    o = op.map_from(m, i)

    for x in range(10):
        result = yield alts([[i, x], o])
        if result.value is None: # write
            print x
        else:
            print "\t", result.value
        # print result



from zope.interface import implements
from collections import namedtuple
from random import shuffle

from csp.impl.interfaces import IHandler
from csp.impl.channels import Box

class AltHandler:
    implements(IHandler)

    def __init__(self, flag, f):
        self.f = f
        self.flag = flag

    def is_active(self):
        return self.flag[0]

    def commit(self):
        self.flag[0] = False
        return self.f


AltResult = namedtuple("AltResult", ["value", "channel"])


# TODO: Support options
def do_alts(operations, callback):
    # XXX Hmm
    assert len(operations) > 0
    flag = [True]
    operations = list(operations)
    # TODO: Accept a priority function or something
    shuffle(operations)

    # XXX: Python uses function-scope not block-scope. Therefore
    # "port" is mutably shared by the iterations. The "lambda port:" trick
    # is used to pass the values of "port" to "callback", instead of
    # passing the mutable reference.
    for operation in operations:
        if isinstance(operation, (list, tuple)):
            port, value = operation
            result = port.put(value, (
                lambda port: AltHandler(flag, lambda ok: callback(AltResult(ok, port)))
            )(port))
        else:
            port = operation
            result = port.take((
                lambda port: AltHandler(flag, lambda value: callback(AltResult(value, port)))
            )(port))
        if result:
            assert isinstance(result, Box)
            return Box(AltResult(result.value, port))



from zope.interface import implements
from collections import deque

from csp.impl.interfaces import IBuffer

# TODO: Check the performance characteristics. A hand-rolled
# linked-list based ring buffer may be better.

class RingBuffer:

    def __init__(self, size, iterable=()):
        self.ring = deque(iterable, size)

    def pop(self):
        """Returns the oldest item, removing it from the buffer.
        """
        return self.ring.popleft()

    def unshift(self, item):
        """Adds 1 item, dropping oldest item to make place if needed.
        """
        self.ring.append(item)

    def unbounded_unshift(self, item):
        """Adds 1 item, extending the buffer to make place if needed.
        """
        if len(self.ring) == self.ring.maxlen:
            self.resize()
        self.unshift(item)

    def resize(self):
        """Doubles the size of the buffer.
        """
        self.ring = deque(self.ring, self.ring.maxlen * 2)

    def cleanup(self, keep):
        """Removes items that do not match the specified predicate.
        """
        i, n = 0, len(self.ring)
        while True:
            i += 1
            if i > n:
                break
            item = self.ring.popleft()
            if keep(item):
                self.ring.append(item)

    def __len__(self):
        return len(self.ring)

    def __repr__(self):
        return "<RingBuffer %s>" % self.ring.__repr__()


class FixedBuffer:
    """Creates a fixed buffer of given size. When full, puts will block.
    """

    implements(IBuffer)

    def __init__(self, size):
        assert size > 0
        self.buf = deque((), size)

    def is_full(self):
        return len(self.buf) == self.buf.maxlen

    def add(self, item):
        assert not self.is_full()
        return self.buf.append(item)

    def remove(self):
        return self.buf.popleft()

    def __len__(self):
        return len(self.buf)

    def __repr__(self):
        return "<FixedBuffer %s>" % self.buf.__repr__()


class DroppingBuffer:
    """Creates a fixed buffer of given size. When full, puts will complete
    but value will be dropped.
    """

    implements(IBuffer)

    def __init__(self, size):
        assert size > 0
        self.buf = deque((), size)

    def is_full(self):
        return False

    def add(self, item):
        if len(self.buf) < self.buf.maxlen:
            return self.buf.append(item)

    def remove(self):
        return self.buf.popleft()

    def __len__(self):
        return len(self.buf)

    def __repr__(self):
        return "<DroppingBuffer %s>" % self.buf.__repr__()


class SlidingBuffer:
    """Creates a fixed buffer of given size. When full, puts will complete
    but the oldest values will be dropped to make place.
    """

    implements(IBuffer)

    def __init__(self, size):
        assert size > 0
        self.buf = deque((), size)

    def is_full(self):
        return False

    def add(self, item):
        return self.buf.append(item)

    def remove(self):
        return self.buf.popleft()

    def __len__(self):
        return len(self.buf)

    def __repr__(self):
        return "<SlidingBuffer %s>" % self.buf.__repr__()



# http://talks.golang.org/2013/advconc.slide#6

from csp import Channel, put, take, go, wait


class Ball:
    hits = 0


def player(name, table):
    while True:
        ball = yield take(table)
        ball.hits += 1
        print name, ball.hits
        yield wait(0.1)
        yield put(table, ball)


def main():
    table = Channel()

    go(player("ping", table))
    go(player("pong", table))

    yield put(table, Ball())
    yield wait(1)



from csp import Channel, close, take

import sys

def main():
    chan = Channel()
    close(chan)

    print (yield take(chan))

    # This would blows up the stack if the implementation is in
    # correct (e.g. lack of dispatch.run in important places (see
    # Process._continue))
    limit = sys.getrecursionlimit()
    count = 0
    while True:
        count += 1
        yield take(chan)
        if count > limit:
            print "Did not blow the stack"
            break



from csp.impl.channels import Box

def map_from(f, ch):
    class Channel:
        def close(self):
            ch.close()

        def put(self, value, handler):
            return ch.put(value, handler)

        def take(self, handler):
            class Handler:
                def is_active(self):
                    return handler.is_active()

                def commit(self):
                    f1 = handler.commit()
                    return lambda v: f1(None if v is None else f(v))

            result = ch.take(Handler())
            if result and result.value is not None:
                return Box(f(result.value))
            else:
                return None

    return Channel()

# def map_from1(f, ch):
#     return MapFrom(f, ch)


# class DerivedChannel:
#     pass

# class DerivedHandler:
#     pass

# def map_from(f, ch):
#     out = DerivedChannel()

#     def take(self, handler):
#         h = DerivedHandler()
#         h.is_active = lambda self: handler.is_active()
#         h.commit = lambda self: lambda v: handler.commit()(
#             None if v is None else f(v))

#         result = ch.take(h)
#         if result and result.value:
#             return Box(f(result.value))
#         else:
#             return None

#     out.close = lambda self: ch.close()
#     out.put = lambda self, value, handler: ch.put(value, handler)
#     out.take = take

#     return out
